package za.co.tcg.tcg_location

import android.content.Context
import android.content.Intent
import android.os.Handler
import android.util.Log
import androidx.core.app.JobIntentService
import com.google.android.gms.location.LocationResult
import io.flutter.embedding.engine.FlutterEngine
import io.flutter.embedding.engine.dart.DartExecutor.DartCallback
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.PluginRegistrantCallback
import io.flutter.view.FlutterCallbackInformation
import io.flutter.view.FlutterMain
import java.util.*
import java.util.concurrent.atomic.AtomicBoolean


class LocationService : MethodCallHandler, JobIntentService() {

    private val queue = ArrayDeque<Pair<Long, String>>()
    private lateinit var mBackgroundChannel: MethodChannel
    private lateinit var mContext: Context

    companion object {
        @JvmStatic
        private val TAG = "LocationService"
        @JvmStatic
        private val JOB_ID = UUID.randomUUID().mostSignificantBits.toInt()
        @JvmStatic
        private var sBackgroundFlutterEngine: FlutterEngine? = null
        @JvmStatic
        private val sServiceStarted = AtomicBoolean(false)

        @JvmStatic
        private lateinit var sPluginRegistrantCallback: PluginRegistrantCallback

        @JvmStatic
        fun enqueueWork(context: Context, work: Intent) {
            enqueueWork(context, LocationService::class.java, JOB_ID, work)
        }

        @JvmStatic
        fun setPluginRegistrant(callback: PluginRegistrantCallback) {
            sPluginRegistrantCallback = callback
        }
    }

    private fun startLocationService(context: Context) {

        synchronized(sServiceStarted) {

            mContext = context

            if (sBackgroundFlutterEngine == null) {

                val callbackHandle = context.getSharedPreferences(
                        TcgLocationPlugin.SHARED_PREFERENCES_KEY,
                        Context.MODE_PRIVATE)
                        .getLong(TcgLocationPlugin.CALLBACK_DISPATCHER_HANDLE_KEY, 0)

                if (callbackHandle == 0L) {
                    Log.e(TAG, "Fatal: no callback registered")
                    return
                }

                val callbackInfo = FlutterCallbackInformation.lookupCallbackInformation(callbackHandle)
                if (callbackInfo == null) {
                    Log.e(TAG, "Fatal: failed to find callback")
                    return
                }
                Log.i(TAG, "Starting LocationService...")
                sBackgroundFlutterEngine = FlutterEngine(context)

                val args = DartCallback(
                        context.getAssets(),
                        FlutterMain.findAppBundlePath(context)!!,
                        callbackInfo
                )
                sBackgroundFlutterEngine!!.getDartExecutor().executeDartCallback(args)
                IsolateHolderService.setBackgroundFlutterEngine(sBackgroundFlutterEngine)
            }
        }
        mBackgroundChannel = MethodChannel(sBackgroundFlutterEngine!!.getDartExecutor().getBinaryMessenger(),
                "tcg_location_background")
        mBackgroundChannel.setMethodCallHandler(this)
    }

    override fun onMethodCall(call: MethodCall, result: Result) {
        when(call.method) {
            "initialized" -> {

                val msg = "Location Service initialized method invoked"
                Log.w(TAG, msg)

                synchronized(sServiceStarted) {
                    while (!queue.isEmpty()) {

                        val pair: Pair<Long, String> = queue.remove()

                        val msg2 = "About to emit pair: ${pair.first}, ${pair.second}"
                        Log.w(TAG, msg2)

                        mBackgroundChannel.invokeMethod("", listOf(pair.first, pair.second))
                    }

                    Log.w(TAG, "About to set sServiceStarted to true")
                    sServiceStarted.set(true)
                }
            }
            "promoteToForeground" -> {
                mContext.startService(Intent(mContext, IsolateHolderService::class.java))
            }
            "demoteToBackground" -> {
                val intent = Intent(mContext, IsolateHolderService::class.java)
                intent.setAction(IsolateHolderService.ACTION_SHUTDOWN)
                mContext.startService(Intent(mContext, IsolateHolderService::class.java))
            }
            else -> result.notImplemented()
        }
        result.success(null)
    }

    override fun onCreate() {
        super.onCreate()
        startLocationService(this)
    }

    override fun onHandleWork(intent: Intent) {

        val p = mContext.getSharedPreferences(TcgLocationPlugin.SHARED_PREFERENCES_KEY, Context.MODE_PRIVATE)
        val callbackHandle = p.getLong(TcgLocationPlugin.CALLBACK_HANDLE_KEY, 0)

        val msg = "onHandleWork invoked with callbackHandle: $callbackHandle \nand intent: ${intent.toString()} \nwith extras: ${intent.extras}"
        Log.w(TAG, msg)

        if (LocationResult.hasResult(intent)) {

            val locationResult = LocationResult.extractResult(intent)

            val lastLocation = locationResult.lastLocation

            synchronized(sServiceStarted) {
                if (!sServiceStarted.get()) {
                    // Queue up geofencing events while background isolate is starting
                    val msg2 = "About to add pair to queue: ${callbackHandle}, ${lastLocation.latitude}|${lastLocation.longitude}"
                    Log.w(TAG, msg2)
                    queue.add(Pair(callbackHandle, "${lastLocation.latitude}|${lastLocation.longitude}"))
                } else {
                    // Callback method name is intentionally left blank.
                    Handler(mContext.mainLooper).post { mBackgroundChannel.invokeMethod("", listOf(callbackHandle, "${lastLocation.latitude}|${lastLocation.longitude}")) }
                }
            }
        } else {

            Log.e(TAG, "Location error: No result")
            return
        }
    }
}
