package za.co.tcg.tcg_location

import androidx.annotation.NonNull;

import android.Manifest
import android.app.Activity
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.os.Build
import android.os.Looper
import android.util.Log
import com.google.android.gms.location.*
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.Registrar

/** TcgLocationPlugin */
public class TcgLocationPlugin: ActivityAware, FlutterPlugin, MethodCallHandler {

  private var mLocationClient: FusedLocationProviderClient? = null
  private lateinit var channel : MethodChannel
  private var mContext : Context? = null
  private var mActivity : Activity? = null

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    mContext = flutterPluginBinding.applicationContext

    mLocationClient = LocationServices.getFusedLocationProviderClient(mContext!!)

    val isNull = mLocationClient == null
    Log.d(TAG, "Initialized location client - isNull: $isNull")

    channel = MethodChannel(flutterPluginBinding.flutterEngine.dartExecutor, "tcg_location")
    channel.setMethodCallHandler(this)
  }

  companion object {
    @JvmStatic
    fun registerWith(registrar: Registrar) {
      val channel = MethodChannel(registrar.messenger(), "tcg_location")
      channel.setMethodCallHandler(TcgLocationPlugin())
    }

    @JvmStatic
    private val TAG = "LocationPlugin"
    @JvmStatic
    val SHARED_PREFERENCES_KEY = "Location_plugin_cache"
    @JvmStatic
    val CALLBACK_HANDLE_KEY = "callback_handle"
    @JvmStatic
    val CALLBACK_DISPATCHER_HANDLE_KEY = "callback_dispatch_handler"
    @JvmStatic
    private val REQUIRED_PERMISSIONS = arrayOf(Manifest.permission.ACCESS_FINE_LOCATION)
    @JvmStatic
    private val sLocationCacheLock = Object()
    @JvmStatic
    private val mLocationRequest: LocationRequest = LocationRequest().apply {
      interval = 15 * 1000  /* 15 secs */
      fastestInterval = 10 * 1000 /* 10 secs */
      priority = LocationRequest.PRIORITY_HIGH_ACCURACY
    }
    @JvmStatic
    fun reRegisterAfterReboot(context: Context) {
      synchronized(sLocationCacheLock) {
        val p = context.getSharedPreferences(SHARED_PREFERENCES_KEY, Context.MODE_PRIVATE)

        initializeService(context, arrayListOf(p.getLong(CALLBACK_DISPATCHER_HANDLE_KEY, 0)))
      }
    }
    @JvmStatic
    private fun initializeService(context: Context, args: ArrayList<*>?) {

      Log.d(TAG, "Initializing LocationService")
      val callbackHandle = args!![0] as Long

      context.getSharedPreferences(SHARED_PREFERENCES_KEY, Context.MODE_PRIVATE)
              .edit()
              .putLong(CALLBACK_DISPATCHER_HANDLE_KEY, callbackHandle)
              .apply()
    }

    @JvmStatic
    private fun startLocation(context: Context, args: ArrayList<*>?, locationClient: FusedLocationProviderClient?): Boolean {

      if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && (context.checkSelfPermission(Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_DENIED)) {
        val msg = "'startLocation' requires the ACCESS_FINE_LOCATION permission."
        Log.w(TAG, msg)
        return false
      }

      val callbackHandle = args!![0] as Long

      context.getSharedPreferences(SHARED_PREFERENCES_KEY, Context.MODE_PRIVATE)
              .edit()
              .putLong(CALLBACK_HANDLE_KEY, callbackHandle)
              .apply()

      Looper.getMainLooper().run {

        locationClient?.requestLocationUpdates(mLocationRequest, getLocationPendingIndent(context, callbackHandle))?.run {
          addOnSuccessListener {

            Log.w(TAG, "Successfully requested location updates with client")
          }
          addOnFailureListener { ex ->
            Log.w(TAG, ex.stackTrace.toString())
            ex.printStackTrace()
          }
        }
      }

      return true
    }

    @JvmStatic
    private fun stopLocation(context: Context, locationClient: FusedLocationProviderClient?): Boolean {

      val p = context.getSharedPreferences(SHARED_PREFERENCES_KEY, Context.MODE_PRIVATE)

      val msg = "Stopping Location Updates"
      Log.w(TAG, msg)

      locationClient?.removeLocationUpdates(getLocationPendingIndent(context, p.getLong(CALLBACK_HANDLE_KEY, 0)))

      return true
    }

    @JvmStatic
    private fun getLocationPendingIndent(context: Context, callbackHandle: Long): PendingIntent {

      val intent = Intent(context, LocationBroadcastReceiver::class.java)
      return PendingIntent.getBroadcast(context, 0, intent, PendingIntent.FLAG_UPDATE_CURRENT)
    }
  }

  override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {

    val args = call.arguments<ArrayList<*>>()

    when(call.method) {
      "getPlatformVersion" -> {
        result.success("Android ${Build.VERSION.RELEASE}")
      }
      "initialize" -> {

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
          mActivity?.requestPermissions(REQUIRED_PERMISSIONS, 12312)
        }
        initializeService(mContext!!, args)
        result.success(true)
      }
      "startLocation" -> {
        result.success(startLocation(mContext!!, args, mLocationClient))
      }
      "stopLocation" -> {
        result.success(stopLocation(mContext!!, mLocationClient))
      }
      "getSampleLocation" -> {
        result.success("52.403435|4.804436")
      }
      "initialized" -> {
        result.success(true)
      }
      "didLaunchFromBackground" -> {
        result.success(false)
      }
      else -> {
        result.notImplemented()
      }
    }
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    channel.setMethodCallHandler(null)
    mContext = null
    mLocationClient = null
  }

  override fun onAttachedToActivity(binding: ActivityPluginBinding) {
    mActivity = binding.activity
  }

  override fun onDetachedFromActivity() {
    mActivity = null
  }

  override fun onDetachedFromActivityForConfigChanges() {
    mActivity = null
  }

  override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
    mActivity = binding.activity
  }

}
