import Flutter
import UIKit
import CoreLocation

public class SwiftTcgLocationPlugin: NSObject, FlutterPlugin {
    
   var _callbackChannel: FlutterMethodChannel?
   var _headlessRunner: FlutterEngine?
   var _callbackMapping: String = "location_callback_mapping"
   var _registrar: FlutterPluginRegistrar?
   var _registerPlugins: FlutterPluginRegistrantCallback?
   var initialized: Bool = false
   var backgroundIsolateRun: Bool = false
    
    public init(registrar: FlutterPluginRegistrar) {
        super.init()
        
        let channel = FlutterMethodChannel(name: "tcg_location", binaryMessenger: registrar.messenger())
        _headlessRunner = FlutterEngine(name: "LocationIsolate", project: nil, allowHeadlessExecution: true)
        
        if let runner = _headlessRunner as? FlutterBinaryMessenger {
            _callbackChannel = FlutterMethodChannel(name: "tcg_location_background", binaryMessenger: runner)
        }
        
        _registrar = registrar
        
        registrar.addMethodCallDelegate(self, channel: channel)
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
       let instance = SwiftTcgLocationPlugin(registrar: registrar)
       registrar.addApplicationDelegate(instance)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
            
        print("Method call invoked: \(call.method)")
        
        switch call.method {
        case "initialize":
            // Init
            result(initialize((call.arguments as! [Any?]).first as! CLongLong))
        case "startLocation":
            result(initLocation((call.arguments as! [Any?]).first as! CLongLong))
        case "stopLocation":
            result(stopLocation())
        case "getSampleLocation":
            result("52.403435|4.804436")
        case "getPlatformVersion":
            result("iOS " + UIDevice.current.systemVersion)
        case "initialized":
            result(true)
        case "didLaunchFromBackground":
            result(didLaunchFromBackground())
        default:
            result(FlutterMethodNotImplemented)
        }
    }
    
    public func didLaunchFromBackground() -> String {
            
        if let didLaunch = UserDefaults.standard.object(forKey: "didLaunchFromBackground") as? Bool {
            
            if didLaunch {
                return "Did launch from background"
            } else {
                return "Did not launch from background"
            }
        }
        
        return "Key not set"
    }

    public func application(_ application: UIApplication, didFinishLaunchingWithOptions launchOptions: [AnyHashable : Any] = [:]) -> Bool {
            
        if (launchOptions[UIApplication.LaunchOptionsKey.location] != nil) {
            

            UserDefaults.standard.set(true, forKey: "didLaunchFromBackground")
            UserDefaults.standard.synchronize()
            /*
            if #available(iOS 10.0, *) {
                let content = UILocalNotification()
                
                content.alertTitle = "iOS Location"
                content.alertBody = "App received location key, initializing"
                content.soundName = UILocalNotificationDefaultSoundName
                
                UIApplication.shared.scheduleLocalNotification(content)
            } else {
                // Fallback on earlier versions
            }
            */
            
            let _ = initialize(getCallbackDispatchHandle())
        }
        
        return true
    }

    // MARK: - Custom Location Code
    private lazy var locationManager: CLLocationManager = {
      let manager = CLLocationManager()
      //manager.desiredAccuracy = kCLLocationAccuracyBest
      manager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters
      manager.delegate = self
      manager.requestAlwaysAuthorization()
        
        if #available(iOS 9.0, *) {
            manager.allowsBackgroundLocationUpdates = true
        } else {
            // Fallback on earlier versions
        }
        
      return manager
        
    }()
    
    public func initialize(_ handle: CLongLong) -> Bool {
        
        setCallbackDispatchHandle(handle)
        
        if let info = FlutterCallbackCache.lookupCallbackInformation(handle) {
            let entryPoint = info.callbackName
            let uri = info.callbackLibraryPath
            
            _headlessRunner?.run(withEntrypoint: entryPoint, libraryURI: uri)
        }
        
        if (!backgroundIsolateRun) {
            
            if let runner = _headlessRunner {
                _registerPlugins?(runner)
            }
        }
        
        if let channel = _callbackChannel {
            _registrar?.addMethodCallDelegate(self, channel: channel)
        }
        
        backgroundIsolateRun = true
        
        return true
    }
    
    public func initLocation(_ handle: CLongLong) -> Bool {
        
        setCallbackHandleForLocationUpdate(handle)
        
        locationManager.startUpdatingLocation()
        
        return true
    }

    public func stopLocation() -> Bool {
        
        locationManager.stopUpdatingLocation()
            
        return true
    }
}

// MARK: - Local State Methods
extension SwiftTcgLocationPlugin {
    
    func setCallbackHandleForLocationUpdate(_ handle: CLongLong) {
        UserDefaults.standard.set(handle, forKey: "location_callback_handle")
        UserDefaults.standard.synchronize()
    }
    
    func getCallbackHandleForLocationUpdate() -> CLongLong {
        if let handle = UserDefaults.standard.object(forKey: "location_callback_handle") as? CLongLong {
            return handle
        }
        
        return 0
    }
    
    func setCallbackDispatchHandle(_ handle: CLongLong) {
        UserDefaults.standard.set(handle, forKey: "callback_dispatcher_handle")
        UserDefaults.standard.synchronize()
    }
    
    func getCallbackDispatchHandle() -> CLongLong {
        if let handle = UserDefaults.standard.object(forKey: "callback_dispatcher_handle") as? CLongLong {
            return handle
        }
        
        return 0
    }
}

// MARK: - CLLocationManagerDelegate
extension SwiftTcgLocationPlugin: CLLocationManagerDelegate {
    
    public func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]) {
        guard let mostRecentLocation = locations.last else {
          return
        }
        
        let callbackHandle = getCallbackHandleForLocationUpdate()
        _callbackChannel?.invokeMethod("", arguments: [callbackHandle, "\(mostRecentLocation.coordinate.latitude)|\(mostRecentLocation.coordinate.longitude)"])
        
        if UIApplication.shared.applicationState == .active {
            
        } else {
            
        }
    }
    
    public func locationManager(_ manager: CLLocationManager, didFailWithError error: Error) {
        
    }
}

